/**
 * @file receiver.c
 * Image receiver
 * @author Marko Mkel (msmakela@nic.funet.fi)
 */

/*
 * Copyright  2004 Marko Mkel
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>

/** Request a file
 * @param width		image width in pixels
 * @param height	image height in pixels
 * @param quality	image quality (1..100)
 * @param output	output stream
 * @return		zero on success, nonzero on failure
 */
static int
receive_file (unsigned width, unsigned height, unsigned quality,
	      FILE* output)
{
  int c;
  size_t total;
  /** file reception buffer */
  static char buf[16384];

  fprintf (stderr, "JPEG %u %u %u;", width, height, quality);

  printf ("JPEG %u %u %u;\n", width, height, quality);
  if (fflush (stdout)) {
    perror ("writing to server failed");
    return 2;
  }

  total = 0;

  for (;;) {
    size_t size, s;
    /* read the block size */
    if (EOF == (c = getchar ())) {
    eof:
      fputs ("read error from server\n", stderr);
      return 2;
    }
    size = (unsigned) c;
    size <<= 8;
    if (EOF == (c = getchar ()))
      goto eof;
    size |= (unsigned) c;
    /* block size of zero signals end of file */
    if (!size)
      break;
    total += size;
    while (size) {
      s = size;
      if (s > sizeof buf)
	s = sizeof buf;
      if (s != fread (buf, 1, s, stdin))
	goto eof;
      if (s != fwrite (buf, 1, s, output)) {
	perror ("write to file\n");
	return 3;
      }
      size -= s;
    }
  }

  /* read feedback from client */
  while (fgets (buf, sizeof buf, stdin)) {
    fputs (buf, stderr);
    if (buf[strlen (buf) - 1] == '\n')
      break;
  }

  /* return zero only if something was transferred */
  return !total;
}

/** Main program
 * @param argc	number of command line arguments
 * @param argv	command line arguments
 */
int
main (int argc, char** argv)
{
  /** program name */
  const char* prog = *argv;
  /** directory name pattern */
  const char* dirname_pattern = "%Y-%m-%d";
  /** file name pattern */
  const char* filename_pattern = "%H:%M:%S.jpg";
  /** image parameters */
  unsigned width = 320, height = 240, quality = 80;
  /** update interval in seconds */
  unsigned update = 60;
  /** buffer for file name strings */
  static char namebuf[1024];

  /* process the options */
  for (argv++; argc > 1 && **argv == '-'; argv++, argc--) {
    const char* opts = *argv;
    while (*++opts) { /* process all flags */
      switch (*opts) {
      case 'h':
	goto Usage;
      case 'q':
      case 'u':
      case 'x':
      case 'y':
	if (argc <= 2) {
	Missing:
	  fprintf (stderr, "receiver -%c: missing argument\n", *opts);
	  goto Usage;
	}
	else {
	  unsigned long ul;
	  char* r;
	  char* s = *++argv;
	  argc--;
	  ul = strtoul (s, &r, 0);
	  if (!*s || *r || !ul) {
	    fprintf (stderr,
		     "receiver -%c: positive number expected, got %s\n",
		     *opts, *argv);
	    goto Usage;
	  }
	  switch (*opts) {
	  case 'x': width = ul; break;
	  case 'y': height = ul; break;
	  case 'u': update = ul; break;
	  case 'q': quality = ul; break;
	  }
	}
	break;
      case 'f':
	if (argc <= 2)
	  goto Missing;
	argc--; filename_pattern = *++argv;
	break;
      case 'd':
	if (argc <= 2)
	  goto Missing;
	argc--; dirname_pattern = *++argv;
	break;
      }
    }
  }

  if (argc > 1) {
    fputs ("receiver: too many arguments\n", stderr);
  Usage:
    fputs ("receiver - receive JPEG images from camera\n"
	   "Usage: ", stderr);
    fputs (prog, stderr);
    fputs (" [options]\n"
	   "Options: -h: request this help text\n"
	   "         -d: directory name pattern for strftime(3)\n"
	   "         -f: file name pattern for strftime(3)\n"
	   "         -q quality: JPEG image quality (1..100)\n"
	   "         -u: update interval in seconds\n"
	   "         -x width: image width in pixels\n"
	   "         -y height: image height in pixels\n", stderr);
    return 0;
  }

  for (;;) {
    FILE* f;
    time_t before, after;
    const struct tm* tm;
    size_t namelen;

  next:
    time (&before);
    tm = localtime (&before);

    /* write a time stamped log entry to stderr */
    fwrite (namebuf, 1, strftime (namebuf, sizeof namebuf,
				  "%Y-%m-%d %H:%M:%S ", tm),
	    stderr);

    /* create directory for output file, if needed */
    namelen = strftime (namebuf, sizeof namebuf, dirname_pattern, tm);
    if (namelen > (sizeof namebuf) - 2) {
      fputs ("directory name too long\n", stderr);
      return 1;
    }
    namebuf[namelen] = 0;
    if (mkdir (namebuf, 0777) && errno != EEXIST) {
      perror ("mkdir");
      return 1;
    }
    /* create the output file */
    namebuf[namelen++] = '/';
    namelen += strftime (namebuf + namelen, (sizeof namebuf) - namelen,
			 filename_pattern, tm);
    if (namelen > (sizeof namebuf) - 1) {
      fputs ("file name too long\n", stderr);
      return 1;
    }
    namebuf[namelen] = 0;

    if (!(f = fopen (namebuf, "wb"))) {
      perror ("open");
      return 1;
    }

    if (receive_file (width, height, quality, f)) {
      fclose (f);
      return 2;
    }

    fclose (f);

    /* make "img.jpg" point to the newest file, atomically; ignore errors */
    unlink ("img-new.jpg");
    link (namebuf, "img-new.jpg");
    rename ("img-new.jpg", "img.jpg");

    /* sleep until next update */
    time (&after);

    if (after - before < update)
      sleep (update - (after - before));
    goto next;
  }
}
